/* ----------------------------------------------------------------------------

     FUNCTION
        ising_2d_pbc   - Program to simulate the Ising model on a simple square
                         lattice with Metroplis dynamics. Periodic boundary
                         conditions are used. This version is particularly fast
                         for off critical temperatures.

                         The following analysis can be performed:

                         a)    Average magnetisation
                         b)    Susceptibility
                         c)    Cumulant
                         d)    Store magnetisations in an array

                         The function type is (int).

     CALLING SEQUENCE

                        (Input Parameters)
           L            Linear system size. (int)
           lattice      Array which holds the spin variables. Must be
                        declared with size of l x l. (int)
           ttc          Temperature in units of the critical temperature. (float)
           h            Applied magnetic field. (float)
           max_sweeps   Maximum number of monte carlo steps. (int)
           start        Start analysis at mcs=start. (int)
           ana          Analyse every ana Monte Carlo step after mcs=start.
           seed         Seed for the random number generator. (int)
           type         This defines the type of analysis. a)-c) will always
                        be performed. Give
                        type = 0      No additional analysis
                        type = 1      Store magnetisations
                        (int)
           init         Initial lattice configuration.
                        init = 0      Hot start, i.e., random configuration.
                        init = 1      All spins are aligned.
                        (int)

                        (Output Parameters)
           a_mag        Average magnetisation. (double)
           sus          Susceptibility. (double)
           cum          Cumulant. (double)
           m_trace      Array of size (mcsmax-start+1)/ana to hold the trace
                        of the magnetisation. In case type is neither 1 nor
                        3 this array is not referenced but must be declared
                        at least as a float variable. (float)


     ERROR CODES
        The functions returns an integer value with the following meaning

           0   -   Everything is well.
          -1   -   Not enough memory.
          -2   -   Lattice contains illegal values.
           1   -   Inconsistency found with decalarations.

     DEFAULT
        By default the lattice is pre-set.

     REQUIRED ROUTINES
        inr250, r250
     
        (c) Satoru Ozawa and Dieter W. Heermann
        Hitachi, Heidelberg July 1994

     REMARKS
        The critical coupling which we use is J/kTc = ln(1+sqrt(2))/2.

-----------------------------------------------------------------------------*/

/* ==== Constants ==== */

# define WindowSize  800        /* Size in pixle of the graphics      */ 
                                /* window                             */ 
# define dx          0.01       /* height and width, in terms of the  */
                                /* normalize window, to draw          */ 

/*--------------------------------------------------------------------------*/
/*      Function: ising_2d_pbc                                              */
/*--------------------------------------------------------------------------*/

int ising_2d_pbc(L,lattice,ttc,h,max_sweeps,start,ana,seed,type,init,
                  a_mag, sus, cum, m_trace)
   int     L;
   int     *lattice;
   float   ttc;
   float   h;
   int     max_sweeps;
   int     start;
   int     ana;
   int     seed;
   int     type;
   int     init;
   double  *a_mag;
   double  *sus;
   double  *cum;
   float   *m_trace;
  {
    /*---------------------------------------------------------- */
    /*              Declarations                                 */
    /*---------------------------------------------------------- */

    /* ====  Area Declarations ==== */

    int      *im, *ip;
    int      mf[251] ;

    float     t_prob[11];
    float     *random;

    /* ==== General declarations ==== */

    int      i,j;
    int      imx,ipx;
    int      adri;
    int      sweep;
    int      size,lsq;
    int      return_code;
    int      ici, sign, en, ich;
    int      m;

    float     J_kbt, temp;
    float     rlsqr;
    float     tmag;
    double    mag, flmag;
    double    mag4;
    double    rmag;
    float     count;
    float     ex;

    /* ==== Pointer declarations ==== */

    int       *l_ptr;
    float     *rand_ptr;

    /*---------------------------------------------------------- */
    /*                End of declares                            */
    /*---------------------------------------------------------- */

    /*---------------------------------------------------------- */
    /*      Start with setting up procedures                     */
    /*---------------------------------------------------------- */

    /* ==== Prepare by initializing all necessary quantities ==== */

    temp      = ttc / 0.440686794;
    size      = L * L;
    rlsqr     = L * L;
    lsq       = L * L;
    m         = - L * L / 2;
    count     = 0.0;
    mag       = 0.0;
    flmag     = 0.0;
    mag4      = 0.0;
    rmag      = 0.0;

    /* ==== Grep the storage for the periodic boundary      ==== */
    /* ==== conditions.                                     ==== */

    ip = (int *) calloc(L, sizeof(int));
    im = (int *) calloc(L, sizeof(int));
    if ( ip == NULL || im == NULL ) {
        return -1;
    }

    /* ==== Initialise now the periodic boundary conditions ==== */

    for(i=0; i < L; i++) {
        ip[i] = i + 1;
        im[i] = i - 1;
    }
    ip[L-1] = 0;
    im[0]   = L - 1;

    /* ==== Grep the storage for the random numbers.        ==== */

    random = (float *) calloc(L, sizeof(float));
    if ( random == NULL ) {
        return -1;
    }

    /* ==== Perpare the random number generator             ==== */

    seed        = (seed << 2) + 1;
    return_code = init_r250(seed,mf);

    if ( init < 2 ) {
       /* ==== Initialize the first configuration.             ==== */

       m     = - L * L / 2;
       l_ptr = lattice;
       for (i=0; i < size; i++) {
          *l_ptr++ = -9;
       }

       if ( init == 0 ) {
           /* ==== Now make a hot start of it ==== */
           for (i=0; i < L; i++) {
               return_code  = r250(L,random,mf);
               rand_ptr     = random;
               for (j=0; j < L; j++) {
                 if ( *rand_ptr++ > 0.5 ) {
                    ici  = lattice[i*L + j];
                    sign = 2 *  (ici >> 4) + 1 ;
                    lattice[i*L + j] = ici - sign * 11;
                    ich               = - 2 * sign;
                    lattice[i*L + im[j]] += ich;
                    lattice[i*L + ip[j]] += ich;
                    lattice[im[i]*L+ j]     += ich;
                    lattice[ip[i]*L+ j]     += ich;
                    m    = m - sign;
                 } /* end if */
               } /* end j */
           } /* end i */
       } /* end init = 0 */
    } /* end of init in case we must set the config */
    else {
       /* ==== Calculate the magnetisation ==== */

       l_ptr = lattice;
       m     = 0;
       for(i=0; i < size; i++) {
           m += 2 *  (*l_ptr++ >> 4) + 1 ;
       }
       m = m/2;
    } /* finished the initial configuration */


    /* ====   Initialise the transition probabilites ==== */

    for(i=1; i < 10; i += 2)  {
        t_prob[i]   = 1.0;
        t_prob[i+1] = 1.0;

        ex = exp((i-5.0)*2.0/temp - h);
        if (ex > 1 ) t_prob[i] = 1.0/ex;

        ex = exp((i-5.0)*2.0/temp + h);
        if (ex > 1 ) t_prob[i+1] = 1.0/ex;
    }

    /* ==== initialize the graphics ==== */

    sgl_winsize(WindowSize,WindowSize);
    sgl_init();

    /*---------------------------------------------------------- */
    /*                M o n t e  C a r l o                       */
    /*---------------------------------------------------------- */

    for(sweep=1; sweep <= max_sweeps; sweep++) {
        for(i=0; i < L; i++)  {
            imx      = im[i];
            ipx      = ip[i];
            adri     = i * L;
            return_code = r250(L,random,mf);
            rand_ptr = random;
            for(j=0; j < L; j++)  {
                ici  = lattice[adri + j];
                sign = 2 *  (ici >> 4) + 1 ;   /* Careful! */
                en   = ici * sign;
                if ( *rand_ptr++ < t_prob[en] )  {
                  m                 = m - sign;
                  lattice[adri + j] = ici - sign * 11;
                  ich               = - 2 * sign;
                  lattice[adri + im[j]] += ich;
                  lattice[adri + ip[j]] += ich;
                  lattice[imx*L+ j]     += ich;
                  lattice[ipx*L+ j]     += ich;
                } /* end of if */
            } /* end j */
        } /* end i */

        /* ---- Graphics part ---- */
        sgl_color_background(SGL_BLACK);
        for(i=0; i < L; i++)  {
            for(j=0; j < L; j++)  {
                ici  = lattice[i*L+j];
                if ( ici > 0 ) {
                    sgl_color(SGL_WHITE);
                    sgl_rectarea((float) i/ (float) L, (float) j/ (float) L, dx,dx); 
                }
                else {
                    sgl_color(SGL_BLACK);
                    sgl_rectarea((float) i / (float) L, (float) j / (float) L, dx,dx); 
                }
            } /* end j */
        } /* end i */
        sgl_display();

       /*---------------------------------------------------------- */
       /*          S a v e   a n d    A n a l y s i s               */
       /*---------------------------------------------------------- */

       /*  ==== check if we want to calculate some quantities ==== */

       if ( start < sweep ) {
         if ( sweep % ana == 0) {
             count  = count + 1.0;
             mag    = 2.0 * abs(m) / rlsqr;
             tmag   = 2.0 * m / rlsqr;
             rmag  += mag;
             flmag += mag * mag;
             mag4  += (mag * mag) * (mag * mag);
             if ( type == 1 ) { *m_trace++ = tmag; }
         } /* end of ana if */
       } /* end of skip */
    } /* end of sweep loop */

  *sus   = (flmag  - rmag *rmag/count) / count;
  *sus  *= rlsqr;
  *a_mag = rmag / count;
  *cum   = 1.0 - mag4 / ( 3.0 * flmag * flmag);
  return 0;
}
